#ifndef FGGL_MATH_TRIS_H
#define FGGL_MATH_TRIS_H

#include <fggl/math/types.hpp>
#include <fggl/data/model.hpp>

namespace fggl::math {

    using Polygon = std::vector<math::vec2>;
    using PolygonVertex = std::vector<data::Vertex2D>;

    constexpr int POSITIVE = 1;
    constexpr int NEGATIVE = -1;
    constexpr int UNSET = 0;

    /**
     * Put an angle in the range [-PI, PI].
     */
    inline float clampAngle( float radianAngle ) {
        if ( radianAngle <= M_PI ) {
            return radianAngle + M_PI_2;
        } else if ( radianAngle > M_PI ) {
            return radianAngle - M_PI_2;
        } else {
            return radianAngle;
        }
    }

    static void checkSign( float value, int& sign, int& firstSign, int& flips ) {
        if ( value > 0 ) {
            if ( sign == UNSET ) {
                firstSign = POSITIVE;
            } else if ( sign < 0 ) {
                flips++;
            }
        } else if ( value < 0 ) {
            if ( sign == UNSET ) {
                firstSign = NEGATIVE;
            } else if ( sign > 0 ) {
                flips++;
            }
            sign = NEGATIVE;
        }
    }

    /**
     * Check if a polygon is convex.
     *
     * see https://math.stackexchange.com/a/1745427
     */
    bool isConvex(const Polygon& polygon) {
        if ( polygon.size() < 3 ) {
            return false;
        }

        const auto n = polygon.size();

        auto wSign = UNSET;

        auto xSign = UNSET;
        auto xFirstSign = UNSET;
        auto xFlips = 0;

        auto ySign = UNSET;
        auto yFirstSign = UNSET;
        auto yFlips = 0;

        auto curr = polygon[ n - 1 ];
        auto next = polygon[ n ];

        for ( auto& v : polygon ) {
            auto prev = curr;
            curr = next;
            next = v;

            auto before = curr - prev;
            auto after = next - curr;

            checkSign( after.x, xSign, xFirstSign, xFlips );
            if ( xFlips > 2 ) {
                return false;
            }

            checkSign( after.y, ySign, yFirstSign, yFlips );
            if ( yFlips > 2 ) {
                return false;
            }

            auto w = before.x * after.y - after.x * before.y;
            if ( wSign == UNSET && w != 0 ) {
                wSign = w;
            } else if ( wSign > 0 && w < 0 ) {
                return false;
            } else if ( wSign < 0 && w > 0 ){
                return false;
            }
        }

        if ( xSign != UNSET && ( xFirstSign != UNSET ) && ( xSign != xFirstSign ) ) {
            xFlips += 1;
        }
        if ( ySign != UNSET && ( yFirstSign != UNSET ) && ( ySign != yFirstSign ) ) {
            yFlips += 1;
        }
        if ( xFlips != 2 || yFlips != 2 ) {
            return false;
        }

        return true;
    }

    static data::Vertex2D pointToVertex( const math::vec2& point ) {
        return data::Vertex2D{
            .position = point,
            .colour = {1.0f, 1.0f, 1.0f}
        };
    }

    /**
     * Fast Triangulation for convex polygons.
     */
    void fanTriangulation(const Polygon& polygon, data::Mesh2D& mesh) {
        assert(polygon.size() >= 3);

        // add the first two points to the mesh
        auto firstIdx = mesh.add_vertex( pointToVertex(polygon[0]) );
        auto prevIdx = mesh.add_vertex( pointToVertex(polygon[1]) );

        // deal with the indicies
        const auto nTris = polygon.size() - 2;
        for ( auto i=0; i < nTris; i++ ) {
            mesh.add_index( firstIdx );
            mesh.add_index( prevIdx );

            auto currIdx = mesh.add_vertex( pointToVertex(polygon[i + 2]) );
            mesh.add_index( currIdx );
            prevIdx = currIdx;
        }
    }

    void fanTriangulation(const PolygonVertex polygon, data::Mesh2D& mesh) {
        assert(polygon.size() >= 3);

        // add the first two points to the mesh
        auto firstIdx = mesh.add_vertex( polygon[0] );
        auto prevIdx = mesh.add_vertex( polygon[1] );

        // deal with the indicies
        const auto nTris = polygon.size() - 2;
        for ( auto i=0; i < nTris; i++ ) {
            mesh.add_index( firstIdx );
            mesh.add_index( prevIdx );

            auto currIdx = mesh.add_vertex( polygon[i + 2] );
            mesh.add_index( currIdx );
            prevIdx = currIdx;
        }
    }


} // namespace fggl::util

#endif

